# -*- coding: utf-8; -*-

from unittest.mock import patch

from wuttaweb import handler as mod, static
from wuttaweb.forms import Form
from wuttaweb.grids import Grid
from wuttaweb.menus import MenuHandler
from wuttaweb.testing import WebTestCase


class MockMenuHandler(MenuHandler):
    pass


class LegacyMenuHandler(MenuHandler):
    pass


class AnotherMenuHandler(MenuHandler):
    pass


class TestWebHandler(WebTestCase):

    def make_handler(self):
        return mod.WebHandler(self.config)

    def test_get_fanstatic_url(self):
        handler = self.make_handler()

        # default with / root path
        url = handler.get_fanstatic_url(self.request, static.logo)
        self.assertEqual(url, "/fanstatic/wuttaweb_img/logo.png")

        # what about a subpath
        with patch.object(self.request, "script_name", new="/testing"):
            url = handler.get_fanstatic_url(self.request, static.logo)
            self.assertEqual(url, "/testing/fanstatic/wuttaweb_img/logo.png")

    def test_get_favicon_url(self):
        handler = self.make_handler()

        # default
        url = handler.get_favicon_url(self.request)
        self.assertEqual(url, "/fanstatic/wuttaweb_img/favicon.ico")

        # config override
        self.config.setdefault("wuttaweb.favicon_url", "/testing/other.ico")
        url = handler.get_favicon_url(self.request)
        self.assertEqual(url, "/testing/other.ico")

    def test_get_header_logo_url(self):
        handler = self.make_handler()

        # default
        url = handler.get_header_logo_url(self.request)
        self.assertEqual(url, "/fanstatic/wuttaweb_img/favicon.ico")

        # config override
        self.config.setdefault("wuttaweb.header_logo_url", "/testing/header.png")
        url = handler.get_header_logo_url(self.request)
        self.assertEqual(url, "/testing/header.png")

    def test_get_main_logo_url(self):
        handler = self.make_handler()

        # default
        url = handler.get_main_logo_url(self.request)
        self.assertEqual(url, "/fanstatic/wuttaweb_img/logo.png")

        # config override
        self.config.setdefault("wuttaweb.logo_url", "/testing/other.png")
        url = handler.get_main_logo_url(self.request)
        self.assertEqual(url, "/testing/other.png")

    def test_get_menu_handler(self):
        handler = self.make_handler()

        # built-in default
        menus = handler.get_menu_handler()
        self.assertIsInstance(menus, MenuHandler)
        self.assertIs(type(menus), MenuHandler)

        # configured default
        self.config.setdefault(
            "wutta.web.menus.handler.default_spec", "tests.test_handler:MockMenuHandler"
        )
        menus = handler.get_menu_handler()
        self.assertIsInstance(menus, MockMenuHandler)

        # configured handler (legacy)
        self.config.setdefault(
            "wutta.web.menus.handler_spec", "tests.test_handler:LegacyMenuHandler"
        )
        menus = handler.get_menu_handler()
        self.assertIsInstance(menus, LegacyMenuHandler)

        # configued handler (proper)
        self.config.setdefault(
            "wutta.web.menus.handler.spec", "tests.test_handler:AnotherMenuHandler"
        )
        menus = handler.get_menu_handler()
        self.assertIsInstance(menus, AnotherMenuHandler)

    def test_get_menu_handler_specs(self):
        handler = self.make_handler()

        # at least one spec by default
        specs = handler.get_menu_handler_specs()
        self.assertIn("wuttaweb.menus:MenuHandler", specs)

        # caller can specify default as string
        specs = handler.get_menu_handler_specs(
            default="tests.test_handler:MockMenuHandler"
        )
        self.assertIn("wuttaweb.menus:MenuHandler", specs)
        self.assertIn("tests.test_handler:MockMenuHandler", specs)
        self.assertNotIn("tests.test_handler:AnotherMenuHandler", specs)

        # caller can specify default as list
        specs = handler.get_menu_handler_specs(
            default=[
                "tests.test_handler:MockMenuHandler",
                "tests.test_handler:AnotherMenuHandler",
            ]
        )
        self.assertIn("wuttaweb.menus:MenuHandler", specs)
        self.assertIn("tests.test_handler:MockMenuHandler", specs)
        self.assertIn("tests.test_handler:AnotherMenuHandler", specs)

        # default can be configured
        self.config.setdefault(
            "wutta.web.menus.handler.default_spec",
            "tests.test_handler:AnotherMenuHandler",
        )
        specs = handler.get_menu_handler_specs()
        self.assertIn("wuttaweb.menus:MenuHandler", specs)
        self.assertNotIn("tests.test_handler:MockMenuHandler", specs)
        self.assertIn("tests.test_handler:AnotherMenuHandler", specs)

        # the rest come from entry points
        with patch.object(
            mod,
            "load_entry_points",
            return_value={
                "legacy": LegacyMenuHandler,
            },
        ):
            specs = handler.get_menu_handler_specs()
            self.assertNotIn("wuttaweb.menus:MenuHandler", specs)
            self.assertNotIn("tests.test_handler:MockMenuHandler", specs)
            self.assertIn("tests.test_handler:LegacyMenuHandler", specs)
            # nb. this remains from previous config default
            self.assertIn("tests.test_handler:AnotherMenuHandler", specs)

    def test_make_form(self):
        handler = self.make_handler()
        form = handler.make_form(self.request)
        self.assertIsInstance(form, Form)

    def test_make_grid(self):
        handler = self.make_handler()
        grid = handler.make_grid(self.request)
        self.assertIsInstance(grid, Grid)



# -*- coding: utf-8; -*-

from unittest.mock import patch, MagicMock

import pytest

from wuttaweb.db import continuum as mod
from wuttaweb.testing import WebTestCase


class TestWuttaWebContinuumPlugin(WebTestCase):

    def setUp(self):
        if not hasattr(mod, "WuttaWebContinuumPlugin"):
            pytest.skip("test not relevant without sqlalchemy-continuum")
        self.setup_web()

    def make_plugin(self):
        return mod.WuttaWebContinuumPlugin()

    def test_get_remote_addr(self):
        plugin = self.make_plugin()

        # typical request from client IP
        self.request.client_addr = "172.237.145.181"
        self.assertEqual(plugin.get_remote_addr(None, self.session), "172.237.145.181")

        # pretend we have no request; IP will be random string
        # (probably 127.0.0.1 but can't guarentee that..)
        with patch.object(mod, "get_current_request", return_value=None):
            addr = plugin.get_remote_addr(None, self.session)
            self.assertIsInstance(addr, str)

    def test_get_user_id(self):
        plugin = self.make_plugin()

        with patch.object(mod, "get_current_request", return_value=None):
            self.assertIsNone(plugin.get_user_id(None, self.session))

        self.request.user = MagicMock(uuid="some-random-uuid")
        self.assertEqual(plugin.get_user_id(None, self.session), "some-random-uuid")



# -*- coding: utf-8; -*-

import datetime
from unittest.mock import patch, MagicMock

from wuttjamaican.reports import Report

import colander
from pyramid.httpexceptions import HTTPNotFound

from wuttaweb.views import reports as mod
from wuttaweb.testing import WebTestCase


class SomeRandomReport(Report):
    """
    This report shows something random.
    """

    report_key = "testing_some_random"
    report_title = "Random Test Report"

    def add_params(self, schema):

        schema.add(
            colander.SchemaNode(colander.String(), name="foo", missing=colander.null)
        )

        schema.add(
            colander.SchemaNode(
                colander.Date(), name="start_date", missing=colander.null
            )
        )

    def get_output_columns(self):
        return ["foo"]

    def make_data(self, params, **kwargs):
        return {
            "output_title": "Testing Output",
            "data": [{"foo": "bar"}],
        }


class TestReportViews(WebTestCase):

    def make_view(self):
        return mod.ReportView(self.request)

    def test_includeme(self):
        self.pyramid_config.include("wuttaweb.views.reports")

    def test_get_grid_data(self):
        view = self.make_view()
        providers = dict(self.app.providers)
        providers["wuttatest"] = MagicMock(report_modules=["tests.views.test_reports"])
        with patch.object(self.app, "providers", new=providers):

            data = view.get_grid_data()
            self.assertIsInstance(data, list)
            self.assertTrue(data)  # 1+ reports

    def test_normalize_report(self):
        view = self.make_view()
        report = SomeRandomReport(self.config)
        normal = view.normalize_report(report)
        help_text = normal.pop("help_text").strip()
        self.assertEqual(help_text, "This report shows something random.")
        self.assertEqual(
            normal,
            {
                "report_key": "testing_some_random",
                "report_title": "Random Test Report",
            },
        )

    def test_configure_grid(self):
        view = self.make_view()
        grid = view.make_model_grid()
        self.assertIn("report_title", grid.searchable_columns)
        self.assertIn("help_text", grid.searchable_columns)

    def test_get_instance(self):
        view = self.make_view()
        providers = {
            "wuttatest": MagicMock(report_modules=["tests.views.test_reports"]),
        }
        with patch.object(self.app, "providers", new=providers):

            # normal
            with patch.object(
                self.request, "matchdict", new={"report_key": "testing_some_random"}
            ):
                report = view.get_instance()
                self.assertIsInstance(report, dict)
                self.assertEqual(report["report_key"], "testing_some_random")
                self.assertEqual(report["report_title"], "Random Test Report")

            # not found
            with patch.object(
                self.request, "matchdict", new={"report_key": "this-should_notEXIST"}
            ):
                self.assertRaises(HTTPNotFound, view.get_instance)

    def test_get_instance_title(self):
        view = self.make_view()
        result = view.get_instance_title({"report_title": "whatever"})
        self.assertEqual(result, "whatever")

    def test_view(self):
        self.pyramid_config.add_route("home", "/")
        self.pyramid_config.add_route("login", "/auth/login")
        self.pyramid_config.add_route("reports", "/reports/")
        self.pyramid_config.add_route("reports.view", "/reports/{report_key}")
        view = self.make_view()
        providers = dict(self.app.providers)
        providers["wuttatest"] = MagicMock(report_modules=["tests.views.test_reports"])
        with patch.object(self.app, "providers", new=providers):
            with patch.object(
                self.request, "matchdict", new={"report_key": "testing_some_random"}
            ):

                # initial view
                response = view.view()
                self.assertEqual(response.status_code, 200)
                # nb. there's a button in there somewhere, but no output title
                self.assertIn("Run Report", response.text)
                self.assertNotIn("Testing Output", response.text)

                # run the report
                with patch.object(
                    self.request,
                    "GET",
                    new={
                        "__start__": "start_date:mapping",
                        "date": "2025-01-11",
                        "__end__": "start_date",
                    },
                ):
                    response = view.view()
                    self.assertEqual(response.status_code, 200)
                    # nb. there's a button in there somewhere, *and* an output title
                    self.assertIn("Run Report", response.text)
                    self.assertIn("Testing Output", response.text)

    def test_configure_form(self):
        view = self.make_view()
        providers = dict(self.app.providers)
        providers["wuttatest"] = MagicMock(report_modules=["tests.views.test_reports"])
        with patch.object(self.app, "providers", new=providers):

            with patch.object(
                self.request, "matchdict", new={"report_key": "testing_some_random"}
            ):
                report = view.get_instance()
                form = view.make_model_form(report)
                self.assertIn("help_text", form.readonly_fields)
                self.assertIn("foo", form)

    def test_normalize_columns(self):
        view = self.make_view()

        columns = view.normalize_columns(["foo"])
        self.assertEqual(
            columns,
            [
                {"name": "foo", "label": "foo"},
            ],
        )

        columns = view.normalize_columns([{"name": "foo"}])
        self.assertEqual(
            columns,
            [
                {"name": "foo", "label": "foo"},
            ],
        )

        columns = view.normalize_columns([{"name": "foo", "label": "FOO"}])
        self.assertEqual(
            columns,
            [
                {"name": "foo", "label": "FOO"},
            ],
        )

        columns = view.normalize_columns(
            [{"name": "foo", "label": "FOO", "numeric": True}]
        )
        self.assertEqual(
            columns,
            [
                {"name": "foo", "label": "FOO", "numeric": True},
            ],
        )

    def test_run_report(self):
        view = self.make_view()
        providers = dict(self.app.providers)
        providers["wuttatest"] = MagicMock(report_modules=["tests.views.test_reports"])
        with patch.object(self.app, "providers", new=providers):

            with patch.object(
                self.request, "matchdict", new={"report_key": "testing_some_random"}
            ):
                report = view.report_handler.get_report("testing_some_random")
                normal = view.normalize_report(report)
                form = view.make_model_form(normal)

                # typical
                context = view.run_report(report, {"form": form})
                self.assertEqual(
                    sorted(context["report_params"]), ["foo", "start_date"]
                )
                self.assertEqual(
                    context["report_data"],
                    {
                        "output_title": "Testing Output",
                        "data": [{"foo": "bar"}],
                    },
                )
                self.assertIn("report_generated", context)

                # invalid params
                with patch.object(self.request, "GET", new={"start_date": "NOT_GOOD"}):
                    context = view.run_report(report, {"form": form})
                    self.assertNotIn("report_params", context)
                    self.assertNotIn("report_data", context)
                    self.assertNotIn("report_generated", context)

                # custom formatter
                with patch.object(report, "get_output_columns") as get_output_columns:
                    get_output_columns.return_value = [
                        "foo",
                        {
                            "name": "start_date",
                            "formatter": lambda val: "FORMATTED VALUE",
                        },
                    ]

                    with patch.object(report, "make_data") as make_data:
                        make_data.return_value = [
                            {"foo": "bar", "start_date": datetime.date(2025, 1, 11)},
                        ]

                        context = view.run_report(report, {"form": form})
                        get_output_columns.assert_called_once_with()
                        self.assertEqual(len(context["report_columns"]), 2)
                        self.assertEqual(context["report_columns"][0]["name"], "foo")
                        self.assertEqual(
                            context["report_columns"][1]["name"], "start_date"
                        )
                        self.assertEqual(
                            context["report_data"],
                            {
                                "output_title": "Random Test Report",
                                "data": [
                                    {"foo": "bar", "start_date": "FORMATTED VALUE"}
                                ],
                            },
                        )

    def test_download_data(self):
        view = self.make_view()
        providers = dict(self.app.providers)
        providers["wuttatest"] = MagicMock(report_modules=["tests.views.test_reports"])
        with patch.object(self.app, "providers", new=providers):
            with patch.object(
                self.request, "matchdict", new={"report_key": "testing_some_random"}
            ):

                params, columns, data = view.get_download_data()
                self.assertEqual(params, {})
                self.assertEqual(columns, [{"name": "foo", "label": "foo"}])
                self.assertEqual(
                    data,
                    {
                        "output_title": "Testing Output",
                        "data": [{"foo": "bar"}],
                    },
                )

    def test_download_path(self):
        view = self.make_view()
        data = {"output_title": "My Report"}
        path = view.get_download_path(data, "csv")
        self.assertTrue(path.endswith("My Report.csv"))



# -*- coding: utf-8; -*-
"""
Tasks for wuttaweb
"""

import os
import shutil

from invoke import task


@task
def release(c, skip_tests=False):
    """
    Release a new version of WuttJamaican
    """
    if not skip_tests:
        c.run("pytest -m 'not versioned and not functional'")
        c.run("pytest -m 'versioned'")
        c.run("pytest -m 'functional'")

    # rebuild pkg
    if os.path.exists("dist"):
        shutil.rmtree("dist")
    if os.path.exists("WuttJamaican.egg-info"):
        shutil.rmtree("WuttJamaican.egg-info")
    c.run("python -m build --sdist")

    # upload
    c.run("twine upload dist/*")



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views of Views
"""

import importlib
import logging
import os
import re
import sys

from mako.lookup import TemplateLookup

from wuttaweb.views import MasterView
from wuttaweb.util import get_model_fields


log = logging.getLogger(__name__)


class MasterViewView(MasterView):  # pylint: disable=abstract-method
    """
    Master view which shows a list of all master views found in the
    app registry.

    Route prefix is ``master_views``; notable URLs provided by this
    class include:

    * ``/views/master/``
    """

    model_name = "master_view"
    model_title = "Master View"
    model_title_plural = "Master Views"
    url_prefix = "/views/master"

    filterable = False
    sortable = True
    sort_on_backend = False
    paginated = True
    paginate_on_backend = False

    creatable = True
    viewable = False  # nb. it has a pseudo-view action instead
    editable = False
    deletable = False
    configurable = True

    labels = {
        "model_title_plural": "Title",
        "url_prefix": "URL Prefix",
    }

    grid_columns = [
        "model_title_plural",
        "model_name",
        "route_prefix",
        "url_prefix",
    ]

    sort_defaults = "model_title_plural"

    def get_grid_data(  # pylint: disable=empty-docstring
        self, columns=None, session=None
    ):
        """ """
        data = []

        # nb. we do not omit any views due to lack of permission here.
        # all views are shown for anyone seeing this page.  this is
        # for sake of clarity so admin users are aware of what is
        # *possible* within the app etc.
        master_views = self.request.registry.settings.get("wuttaweb_master_views", {})
        for model_views in master_views.values():
            for view in model_views:
                data.append(
                    {
                        "model_title_plural": view.get_model_title_plural(),
                        "model_name": view.get_model_name(),
                        "route_prefix": view.get_route_prefix(),
                        "url_prefix": view.get_url_prefix(),
                    }
                )

        return data

    def configure_grid(self, grid):  # pylint: disable=empty-docstring
        """ """
        g = grid
        super().configure_grid(g)

        # nb. show more views by default
        g.pagesize = 50

        # nb. add "pseudo" View action
        def viewurl(view, i):  # pylint: disable=unused-argument
            return self.request.route_url(view["route_prefix"])

        g.add_action("view", icon="eye", url=viewurl)

        # model_title_plural
        g.set_link("model_title_plural")
        g.set_searchable("model_title_plural")

        # model_name
        g.set_searchable("model_name")

        # route_prefix
        g.set_searchable("route_prefix")

        # url_prefix
        g.set_link("url_prefix")
        g.set_searchable("url_prefix")

    def get_template_context(self, context):  # pylint: disable=empty-docstring
        """ """
        if self.creating:
            model = self.app.model
            session = self.Session()

            # app models
            app_models = []
            for name in dir(model):
                obj = getattr(model, name)
                if (
                    isinstance(obj, type)
                    and issubclass(obj, model.Base)
                    and obj is not model.Base
                ):
                    app_models.append(name)
            context["app_models"] = sorted(app_models)

            # view module location
            view_locations = self.get_view_module_options()
            modpath = self.config.get("wuttaweb.master_views.default_module_dir")
            if modpath not in view_locations:
                modpath = None
            if not modpath and len(view_locations) == 1:
                modpath = view_locations[0]
            context["view_module_dirs"] = view_locations
            context["view_module_dir"] = modpath

            # menu handler path
            web = self.app.get_web_handler()
            menu = web.get_menu_handler()
            context["menu_path"] = sys.modules[menu.__class__.__module__].__file__

            # roles for access
            roles = self.get_roles_for_access(session)
            context["roles"] = [
                {"uuid": role.uuid.hex, "name": role.name} for role in roles
            ]
            context["listing_roles"] = {role.uuid.hex: False for role in roles}
            context["creating_roles"] = {role.uuid.hex: False for role in roles}
            context["viewing_roles"] = {role.uuid.hex: False for role in roles}
            context["editing_roles"] = {role.uuid.hex: False for role in roles}
            context["deleting_roles"] = {role.uuid.hex: False for role in roles}

        return context

    def get_roles_for_access(  # pylint: disable=missing-function-docstring
        self, session
    ):
        model = self.app.model
        auth = self.app.get_auth_handler()
        admin = auth.get_role_administrator(session)
        return (
            session.query(model.Role)
            .filter(model.Role.uuid != admin.uuid)
            .order_by(model.Role.name)
            .all()
        )

    def get_view_module_options(self):  # pylint: disable=missing-function-docstring
        modules = set()
        master_views = self.request.registry.settings.get("wuttaweb_master_views", {})
        for model_views in master_views.values():
            for view in model_views:
                parent = ".".join(view.__module__.split(".")[:-1])
                modules.add(parent)
        return sorted(modules)

    def wizard_action(self):  # pylint: disable=too-many-return-statements
        """
        AJAX view to handle various actions for the "new master view" wizard.
        """
        data = self.request.json_body
        action = data.get("action", "").strip()
        try:
            # nb. cannot use match/case statement until python 3.10, but this
            # project technically still supports python 3.8
            if action == "suggest_details":
                return self.suggest_details(data)
            if action == "write_view_file":
                return self.write_view_file(data)
            if action == "check_route":
                return self.check_route(data)
            if action == "apply_permissions":
                return self.apply_permissions(data)
            if action == "":
                return {"error": "Must specify the action to perform."}
            return {"error": f"Unknown action requested: {action}"}

        except Exception as err:  # pylint: disable=broad-exception-caught
            log.exception("new master view wizard action failed: %s", action)
            return {"error": f"Unexpected error occurred: {err}"}

    def suggest_details(  # pylint: disable=missing-function-docstring,too-many-locals
        self, data
    ):
        model = self.app.model
        model_name = data["model_name"]

        def make_normal(match):
            return "_" + match.group(1).lower()

        # normal is like:  poser_widget
        normal = re.sub(r"([A-Z])", make_normal, model_name)
        normal = normal.lstrip("_")

        def make_title(match):
            return " " + match.group(1).upper()

        # title is like:  Poser Widget
        title = re.sub(r"(?:^|_)([a-z])", make_title, normal)
        title = title.lstrip(" ")

        model_title = title
        model_title_plural = title + "s"

        def make_camel(match):
            return match.group(1).upper()

        # camel is like:  PoserWidget
        camel = re.sub(r"(?:^|_)([a-z])", make_camel, normal)

        # fields are unknown without model class
        grid_columns = []
        form_fields = []

        if data["model_option"] == "model_class":
            model_class = getattr(model, model_name)

            # get model title from model class, if possible
            if hasattr(model_class, "__wutta_hint__"):
                model_title = model_class.__wutta_hint__.get("model_title", model_title)
                model_title_plural = model_class.__wutta_hint__.get(
                    "model_title_plural", model_title + "s"
                )

            # get columns/fields from model class
            grid_columns = get_model_fields(self.config, model_class)
            form_fields = grid_columns

        # plural is like:  poser_widgets
        plural = re.sub(r"(?:^| )([A-Z])", make_normal, model_title_plural)
        plural = plural.lstrip("_")

        route_prefix = plural
        url_prefix = "/" + (plural).replace("_", "-")

        return {
            "class_file_name": plural + ".py",
            "class_name": camel + "View",
            "model_name": model_name,
            "model_title": model_title,
            "model_title_plural": model_title_plural,
            "route_prefix": route_prefix,
            "permission_prefix": route_prefix,
            "url_prefix": url_prefix,
            "template_prefix": url_prefix,
            "grid_columns": "\n".join(grid_columns),
            "form_fields": "\n".join(form_fields),
        }

    def write_view_file(self, data):  # pylint: disable=missing-function-docstring
        model = self.app.model

        # sort out the destination file path
        modpath = data["view_location"]
        if modpath:
            mod = importlib.import_module(modpath)
            file_path = os.path.join(
                os.path.dirname(mod.__file__), data["view_file_name"]
            )
        else:
            file_path = data["view_file_path"]

        # confirm file is writable
        if os.path.exists(file_path):
            if data["overwrite"]:
                os.remove(file_path)
            else:
                return {"error": "File already exists"}

        # guess its dotted module path
        modname, ext = os.path.splitext(  # pylint: disable=unused-variable
            os.path.basename(file_path)
        )
        if modpath:
            modpath = f"{modpath}.{modname}"
        else:
            modpath = f"poser.web.views.{modname}"

        # inject module for class if needed
        if data["model_option"] == "model_class":
            model_class = getattr(model, data["model_name"])
            data["model_module"] = model_class.__module__

        # TODO: make templates dir configurable?
        view_templates = TemplateLookup(
            directories=[self.app.resource_path("wuttaweb:code-templates")]
        )

        # render template to file
        template = view_templates.get_template("/new-master-view.mako")
        content = template.render(**data)
        with open(file_path, "wt", encoding="utf_8") as f:
            f.write(content)

        return {
            "view_file_path": file_path,
            "view_module_path": modpath,
            "view_config_path": os.path.join(os.path.dirname(file_path), "__init__.py"),
        }

    def check_route(self, data):  # pylint: disable=missing-function-docstring
        try:
            url = self.request.route_url(data["route"])
            path = self.request.route_path(data["route"])
        except Exception as err:  # pylint: disable=broad-exception-caught
            return {"problem": self.app.render_error(err)}

        return {"url": url, "path": path}

    def apply_permissions(  # pylint: disable=missing-function-docstring,too-many-branches
        self, data
    ):
        session = self.Session()
        auth = self.app.get_auth_handler()
        roles = self.get_roles_for_access(session)
        permission_prefix = data["permission_prefix"]

        if "listing_roles" in data:
            listing = data["listing_roles"]
            for role in roles:
                if listing.get(role.uuid.hex):
                    auth.grant_permission(role, f"{permission_prefix}.list")
                else:
                    auth.revoke_permission(role, f"{permission_prefix}.list")

        if "creating_roles" in data:
            creating = data["creating_roles"]
            for role in roles:
                if creating.get(role.uuid.hex):
                    auth.grant_permission(role, f"{permission_prefix}.create")
                else:
                    auth.revoke_permission(role, f"{permission_prefix}.create")

        if "viewing_roles" in data:
            viewing = data["viewing_roles"]
            for role in roles:
                if viewing.get(role.uuid.hex):
                    auth.grant_permission(role, f"{permission_prefix}.view")
                else:
                    auth.revoke_permission(role, f"{permission_prefix}.view")

        if "editing_roles" in data:
            editing = data["editing_roles"]
            for role in roles:
                if editing.get(role.uuid.hex):
                    auth.grant_permission(role, f"{permission_prefix}.edit")
                else:
                    auth.revoke_permission(role, f"{permission_prefix}.edit")

        if "deleting_roles" in data:
            deleting = data["deleting_roles"]
            for role in roles:
                if deleting.get(role.uuid.hex):
                    auth.grant_permission(role, f"{permission_prefix}.delete")
                else:
                    auth.revoke_permission(role, f"{permission_prefix}.delete")

        return {}

    def configure_get_simple_settings(self):  # pylint: disable=empty-docstring
        """ """
        return [
            {"name": "wuttaweb.master_views.default_module_dir"},
        ]

    def configure_get_context(  # pylint: disable=empty-docstring,arguments-differ
        self, **kwargs
    ):
        """ """
        context = super().configure_get_context(**kwargs)

        context["view_module_locations"] = self.get_view_module_options()

        return context

    @classmethod
    def defaults(cls, config):  # pylint: disable=empty-docstring
        """ """
        cls._masterview_defaults(config)
        cls._defaults(config)

    # pylint: disable=duplicate-code
    @classmethod
    def _masterview_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        permission_prefix = cls.get_permission_prefix()
        model_title_plural = cls.get_model_title_plural()
        url_prefix = cls.get_url_prefix()

        # fix permission group
        config.add_wutta_permission_group(
            permission_prefix, model_title_plural, overwrite=False
        )

        # wizard actions
        config.add_route(
            f"{route_prefix}.wizard_action",
            f"{url_prefix}/new/wizard-action",
            request_method="POST",
        )
        config.add_view(
            cls,
            attr="wizard_action",
            route_name=f"{route_prefix}.wizard_action",
            renderer="json",
            permission=f"{permission_prefix}.create",
        )

    # pylint: enable=duplicate-code


def defaults(config, **kwargs):  # pylint: disable=missing-function-docstring
    base = globals()

    MasterViewView = kwargs.get(  # pylint: disable=invalid-name,redefined-outer-name
        "MasterViewView", base["MasterViewView"]
    )
    MasterViewView.defaults(config)


def includeme(config):  # pylint: disable=missing-function-docstring
    defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Tools for displaying simple data diffs
"""

import sqlalchemy as sa
from sqlalchemy import orm

from pyramid.renderers import render
from webhelpers2.html import HTML

from wuttjamaican.diffs import Diff


class WebDiff(Diff):
    """
    Simple diff class for the web app.

    This is based on the
    :class:`~wuttjamaican:wuttjamaican.diffs.Diff` class; it just
    tweaks :meth:`render_html()` to use the web template lookup
    engine.
    """

    cell_padding = None

    def render_html(self, template="/diff.mako", **kwargs):
        """
        Render the diff as HTML table.

        :param template: Name of template to render, if you need to
           override the default.

        :param \\**kwargs: Remaining kwargs are passed as context to
           the template renderer.

        :returns: HTML literal string
        """
        context = kwargs
        context["diff"] = self
        html = render(template, context)
        return HTML.literal(html)


class VersionDiff(WebDiff):
    """
    Special diff class for use with version history views.  While
    based on :class:`WebDiff`, this class uses a different signature
    for the constructor.

    :param config: The app :term:`config object`.

    :param version: Reference to a Continuum version record object.

    :param \\**kwargs: Remaining kwargs are passed as-is to the
       :class:`WebDiff` constructor.
    """

    def __init__(self, config, version, **kwargs):
        import sqlalchemy_continuum as continuum  # pylint: disable=import-outside-toplevel
        from wutta_continuum.util import (  # pylint: disable=import-outside-toplevel
            render_operation_type,
        )

        self.version = version
        self.model_class = continuum.parent_class(type(self.version))
        self.mapper = sa.inspect(self.model_class)
        self.version_mapper = sa.inspect(type(self.version))
        self.title = kwargs.pop("title", self.model_class.__name__)

        self.operation_title = render_operation_type(self.version.operation_type)

        if "nature" not in kwargs:
            if (
                version.previous
                and version.operation_type == continuum.Operation.DELETE
            ):
                kwargs["nature"] = "delete"
            elif version.previous:
                kwargs["nature"] = "update"
            else:
                kwargs["nature"] = "create"

        if "fields" not in kwargs:
            kwargs["fields"] = self.get_default_fields()

        old_data = {}
        new_data = {}
        for field in kwargs["fields"]:
            if version.previous:
                old_data[field] = getattr(version.previous, field)
            new_data[field] = getattr(version, field)

        super().__init__(config, old_data, new_data, **kwargs)

    def get_default_fields(self):  # pylint: disable=missing-function-docstring
        fields = sorted(self.version_mapper.columns.keys())

        unwanted = [
            "transaction_id",
            "end_transaction_id",
            "operation_type",
        ]

        return [field for field in fields if field not in unwanted]

    def render_version_value(self, version, field, value):
        """
        Render the cell value HTML for a given version + field.

        This method is used to render both sides of the diff (old +
        new values).  It will just render the field value using a
        monospace font by default.  However:

        If the field is involved in a mapper relationship (i.e. it is
        the "foreign key" to a related table), the logic here will
        also (try to) traverse that show display text for the related
        object (if found).

        :param version: Reference to the Continuum version object.

        :param field: Name of the field, as string.

        :param value: Raw value for the field, as obtained from the
           version object.

        :returns: Rendered cell value as HTML literal
        """
        # first render normal span; this is our fallback but also may
        # be embedded within a more complex result.
        text = HTML.tag("span", c=[repr(value)], style="font-family: monospace;")

        # style to apply for bold human-friendly text (if applicable)
        bold = "margin-left: 2rem; font-style: italic; font-weight: bold;"

        # check for standard datetime field
        if self.mapper.has_property(field):
            prop = self.mapper.get_property(field)
            if isinstance(prop, orm.ColumnProperty):
                if len(prop.columns) == 1:
                    col = prop.columns[0]
                    if isinstance(col.type, sa.DateTime):
                        if value:
                            # render as local datetime w/ "time since" tooltip
                            display = HTML.tag(
                                "span",
                                c=self.app.render_datetime(value, html=True),
                                style=bold,
                            )
                            return HTML.tag("span", c=[text, display])

        # loop thru all mapped relationship props
        for prop in self.mapper.relationships:

            # we only want singletons
            if prop.uselist:
                continue

            # loop thru columns for prop
            # nb. there should always be just one colum for a
            # singleton prop, but technically a list is used, so no
            # harm in looping i assume..
            for col in prop.local_columns:

                # we only want the matching column
                if col.name != field:
                    continue

                # grab "related version" reference via prop key.  this
                # would be like a UserVersion for instance.
                if ref := getattr(version, prop.key):

                    # grab "related object" reference.  this would be
                    # like a User for instance.
                    if ref := getattr(ref, "version_parent", None):

                        # render text w/ related object as bold string
                        return HTML.tag(
                            "span",
                            c=[text, HTML.tag("span", c=[str(ref)], style=bold)],
                        )

        return text

    def render_old_value(self, field):
        if self.nature == "create":
            return ""
        value = self.old_value(field)
        return self.render_version_value(self.version.previous, field, value)

    def render_new_value(self, field):
        if self.nature == "delete":
            return ""
        value = self.new_value(field)
        return self.render_version_value(self.version, field, value)



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for Alembic
"""

import datetime
import logging
import os
import re

from alembic import command as alembic_command
from alembic.migration import MigrationContext
from alembic.util import CommandError

from wuttjamaican.db.conf import (
    make_alembic_config,
    get_alembic_scriptdir,
    check_alembic_current,
)

import colander
from webhelpers2.html import tags, HTML

from wuttaweb.views import View, MasterView
from wuttaweb.forms import widgets
from wuttaweb.forms.schema import WuttaDateTime


log = logging.getLogger(__name__)


def normalize_revision(
    config, rev, json_safe=False
):  # pylint: disable=missing-function-docstring
    app = config.get_app()

    created = None
    if match := re.search(r"Create Date: (\d{4}-\d{2}-\d{2}[\d:\. ]+\d)", rev.longdoc):
        created = datetime.datetime.fromisoformat(match.group(1))
        created = app.localtime(created, from_utc=False)
        if json_safe:
            created = app.render_datetime(created)

    return {
        "revision": rev.revision,
        "branch_labels": ", ".join(rev.branch_labels),
        "doc": rev.doc,
        "longdoc": rev.longdoc,
        "path": rev.path,
        "dependencies": rev.dependencies or "",
        "down_revision": rev.down_revision or "",
        "nextrev": ", ".join(rev.nextrev),
        "is_base": app.render_boolean(rev.is_base),
        "is_branch_point": app.render_boolean(rev.is_branch_point),
        "is_head": rev.is_head,
        "created": created,
    }


class AlembicDashboardView(View):
    """
    Custom views for the Alembic Dashboard.
    """

    def dashboard(self):
        """
        Main view for the Alembic Dashboard.

        Route name is ``alembic.dashboard``; URL is
        ``/alembic/dashboard``
        """
        script = get_alembic_scriptdir(self.config)
        with self.config.appdb_engine.begin() as conn:
            context = MigrationContext.configure(conn)
            current_heads = context.get_current_heads()

        def normalize(rev):
            normal = normalize_revision(self.config, rev, json_safe=True)
            normal["is_current"] = rev.revision in current_heads

            normal["revision"] = tags.link_to(
                normal["revision"],
                self.request.route_url(
                    "alembic.migrations.view", revision=normal["revision"]
                ),
            )

            if normal["down_revision"]:
                normal["down_revision"] = tags.link_to(
                    normal["down_revision"],
                    self.request.route_url(
                        "alembic.migrations.view", revision=normal["down_revision"]
                    ),
                )

            return normal

        script_heads = []
        for head in script.get_heads():
            rev = script.get_revision(head)
            script_heads.append(normalize(rev))

        db_heads = []
        for head in current_heads:
            rev = script.get_revision(head)
            db_heads.append(normalize(rev))

        script_heads.sort(key=lambda rev: rev["branch_labels"])
        db_heads.sort(key=lambda rev: rev["branch_labels"])

        return {
            "index_title": "Alembic Dashboard",
            "script": {
                "dir": script.dir,
                "version_locations": sorted(script.version_locations),
                "env_py_location": script.env_py_location,
                "file_template": script.file_template,
            },
            "script_heads": script_heads,
            "db_heads": db_heads,
        }

    def migrate(self):
        """
        Action view to migrate the database.  POST request must be used.

        This directly invokes the :func:`alembic upgrade
        <alembic:alembic.command.upgrade>` (or :func:`alembic
        downgrade <alembic:alembic.command.downgrade>`) command.

        It then sets a flash message per the command status, and
        redirects user back to the Dashboard (or other referrer).

        The request must specify a ``revspec`` param, which we pass
        along as-is to the ``alembic`` command.  We assume ``alembic
        upgrade`` unless the request sets ``direction`` param to
        ``"downgrade"``.
        """
        referrer = self.request.get_referrer(
            default=self.request.route_url("alembic.dashboard")
        )
        if self.request.method != "POST":
            return self.redirect(referrer)

        revspec = self.request.POST.get("revspec")
        if not revspec:
            self.request.session.flash("You must provide a target revspec.", "error")

        else:
            direction = self.request.POST.get("direction")
            if direction != "downgrade":
                direction = "upgrade"
            alembic = make_alembic_config(self.config)
            command = (
                alembic_command.downgrade
                if direction == "downgrade"
                else alembic_command.upgrade
            )

            # invoke alembic upgrade/downgrade
            try:
                command(alembic, revspec)
            except Exception as err:  # pylint: disable=broad-exception-caught
                log.exception(
                    "database failed to %s using revspec: %s", direction, revspec
                )
                self.request.session.flash(f"Migrate failed: {err}", "error")
            else:
                self.request.session.flash("Database has been migrated.")

        return self.redirect(referrer)

    @classmethod
    def defaults(cls, config):  # pylint: disable=empty-docstring
        """ """
        cls._defaults(config)

    @classmethod
    def _defaults(cls, config):

        # permission group
        config.add_wutta_permission_group(
            "alembic", "Alembic (General)", overwrite=False
        )

        # dashboard
        config.add_wutta_permission(
            "alembic",
            "alembic.dashboard",
            "Basic (view) access to the Alembic Dashboard",
        )
        config.add_route("alembic.dashboard", "/alembic/dashboard")
        config.add_view(
            cls,
            attr="dashboard",
            route_name="alembic.dashboard",
            renderer="/alembic/dashboard.mako",
            permission="alembic.dashboard",
        )

        # migrate
        config.add_wutta_permission(
            "alembic",
            "alembic.migrate",
            "Run migration scripts on the database",
        )
        config.add_route("alembic.migrate", "/alembic/migrate")
        config.add_view(
            cls,
            attr="migrate",
            route_name="alembic.migrate",
            permission="alembic.migrate",
        )


class AlembicMigrationView(MasterView):  # pylint: disable=abstract-method
    """
    Master view for Alembic Migrations.

    Route prefix is ``alembic.migrations``; notable URLs include:

    * ``/alembic/migrations/``
    * ``/alembic/migrations/new``
    * ``/alembic/migrations/XXX``
    """

    # pylint: disable=duplicate-code
    model_name = "alembic_migration"
    model_key = "revision"
    model_title = "Alembic Migration"
    route_prefix = "alembic.migrations"
    url_prefix = "/alembic/migrations"
    filterable = False
    sortable = True
    sort_on_backend = False
    paginated = True
    paginate_on_backend = False
    editable = False
    configurable = True
    # pylint: enable=duplicate-code

    labels = {
        "doc": "Description",
        "longdoc": "Long Description",
        "nextrev": "Next Revision",
    }

    grid_columns = [
        "is_head",
        "revision",
        "doc",
        "branch_labels",
        "down_revision",
        "created",
    ]

    sort_defaults = ("is_head", "desc")

    form_fields = [
        "revision",
        "doc",
        "longdoc",
        "branch_labels",
        "dependencies",
        "down_revision",
        "nextrev",
        "is_base",
        "is_branch_point",
        "is_head",
        "path",
        "created",
    ]

    def get_grid_data(  # pylint: disable=empty-docstring
        self, columns=None, session=None
    ):
        """ """
        data = []
        script = get_alembic_scriptdir(self.config)
        for rev in script.walk_revisions():
            data.append(normalize_revision(self.config, rev))
        return data

    def configure_grid(self, grid):  # pylint: disable=empty-docstring
        """ """
        g = grid
        super().configure_grid(g)

        # revision
        g.set_link("revision")
        g.set_searchable("revision")

        # doc
        g.set_link("doc")
        g.set_searchable("doc")

        # branch_labels
        g.set_searchable("branch_labels")

        # is_head
        g.set_label("is_head", "Head")
        g.set_renderer("is_head", self.render_is_head)

        # created
        g.set_renderer("created", "datetime")

    def render_is_head(  # pylint: disable=missing-function-docstring,unused-argument
        self, rev, field, value
    ):
        return self.app.render_boolean(value) if value else ""

    def get_instance(
        self, **kwargs
    ):  # pylint: disable=empty-docstring,arguments-differ,unused-argument
        """ """
        if "_cached_instance" not in self.__dict__:
            revision = self.request.matchdict["revision"]
            script = get_alembic_scriptdir(self.config)
            try:
                rev = script.get_revision(revision)
            except CommandError:
                rev = None
            if not rev:
                raise self.notfound()
            self.__dict__["_cached_instance"] = normalize_revision(self.config, rev)
        return self.__dict__["_cached_instance"]

    def get_instance_title(self, instance):  # pylint: disable=empty-docstring
        """ """
        text = f"({instance['branch_labels']}) {instance['doc']}"
        if instance.get("is_head"):
            text += " [head]"
        return text

    def configure_form(self, form):  # pylint: disable=empty-docstring
        """ """
        f = form
        super().configure_form(f)

        # revision
        f.set_widget("revision", widgets.CopyableTextWidget())

        # longdoc
        f.set_widget("longdoc", "notes")

        # down_revision
        f.set_widget("down_revision", widgets.AlembicRevisionWidget(self.request))

        # nextrev
        f.set_widget("nextrev", widgets.AlembicRevisionsWidget(self.request))

        # is_head
        f.set_node("is_head", colander.Boolean())

        # path
        f.set_widget("path", widgets.CopyableTextWidget())

        # created
        f.set_node("created", WuttaDateTime())
        f.set_widget("created", widgets.WuttaDateTimeWidget(self.request))

    def make_create_form(self):  # pylint: disable=empty-docstring
        """ """
        alembic = make_alembic_config(self.config)
        script = get_alembic_scriptdir(self.config, alembic)

        schema = colander.Schema()
        schema.add(colander.SchemaNode(colander.String(), name="description"))

        schema.add(
            colander.SchemaNode(
                colander.Boolean(),
                name="autogenerate",
                default=check_alembic_current(self.config, alembic),
            )
        )

        schema.add(
            colander.SchemaNode(
                colander.String(), name="branching_option", default="revise"
            )
        )

        branch_options = self.get_revise_branch_options(script)

        revise_branch = colander.SchemaNode(
            colander.String(),
            name="revise_branch",
            missing=colander.null,
            validator=colander.OneOf(branch_options),
            widget=widgets.SelectWidget(values=[(b, b) for b in branch_options]),
        )

        branch = self.config.get(f"{self.config.appname}.alembic.default_revise_branch")
        if not branch and len(branch_options) == 1:
            branch = branch_options[0]
        if branch:
            revise_branch.default = branch

        schema.add(revise_branch)

        schema.add(
            colander.SchemaNode(
                colander.String(), name="new_branch", missing=colander.null
            )
        )

        version_locations = sorted(
            self.config.parse_list(alembic.get_main_option("version_locations"))
        )

        schema.add(
            colander.SchemaNode(
                colander.String(),
                name="version_location",
                missing=colander.null,
                validator=colander.OneOf(version_locations),
                widget=widgets.SelectWidget(values=[(v, v) for v in version_locations]),
            )
        )

        schema.validator = colander.All(
            self.validate_revise_branch, self.validate_new_branch
        )

        form = self.make_form(
            schema=schema,
            cancel_url_fallback=self.get_index_url(),
            button_label_submit="Write Script File",
        )

        form.set_label("revise_branch", "Branch")

        return form

    def validate_revise_branch(  # pylint: disable=missing-function-docstring
        self, node, value
    ):
        if value["branching_option"] == "revise":
            if not value["revise_branch"]:
                node["revise_branch"].raise_invalid(
                    "Must specify which branch to revise."
                )

    def validate_new_branch(  # pylint: disable=missing-function-docstring
        self, node, value
    ):
        if value["branching_option"] == "new":

            if not value["new_branch"]:
                node["new_branch"].raise_invalid("New branch requires a name.")

            if not value["version_location"]:
                node["version_location"].raise_invalid(
                    "New branch requires a version location."
                )

    def save_create_form(self, form):  # pylint: disable=empty-docstring
        """ """
        alembic = make_alembic_config(self.config)
        data = form.validated

        # kwargs for `alembic revision` command
        kw = {
            "message": data["description"],
            "autogenerate": data["autogenerate"],
        }
        if data["branching_option"] == "new":
            kw["head"] = "base"
            kw["branch_label"] = data["new_branch"]
            kw["version_path"] = self.app.resource_path(data["version_location"])
        else:
            assert data["branching_option"] == "revise"
            kw["head"] = f"{data['revise_branch']}@head"

        # run `alembic revision`
        revision = alembic_command.revision(alembic, **kw)

        intro = HTML.tag(
            "p",
            class_="block",
            c="New migration script has been created.  "
            "Please review and modify the file contents as needed:",
        )

        path = HTML.tag(
            "p",
            class_="block has-background-white has-text-black is-family-monospace",
            style="padding: 0.5rem;",
            c=[HTML.tag("wutta-copyable-text", text=revision.path)],
        )

        outro = HTML.tag(
            "p",
            class_="block",
            c=[
                "When satisfied, proceed to ",
                tags.link_to(
                    "Migrate Database", self.request.route_url("alembic.dashboard")
                ),
                ".",
            ],
        )

        self.request.session.flash(HTML.tag("div", c=[intro, path, outro]))
        return revision

    def save_delete_form(self, form):  # pylint: disable=empty-docstring
        """ """
        rev = self.get_instance()
        os.remove(rev["path"])

    # TODO: this is effectivey duplicated in AppTableView.get_migration_branch_options()
    def get_revise_branch_options(  # pylint: disable=missing-function-docstring
        self, script
    ):
        branches = set()
        for rev in script.get_revisions(script.get_heads()):
            branches.update(rev.branch_labels)
        return sorted(branches)

    def configure_get_simple_settings(self):  # pylint: disable=empty-docstring
        """ """
        return [
            {"name": f"{self.config.appname}.alembic.default_revise_branch"},
        ]

    def configure_get_context(  # pylint: disable=empty-docstring,arguments-differ
        self, **kwargs
    ):
        """ """
        context = super().configure_get_context(**kwargs)

        script = get_alembic_scriptdir(self.config)
        context["revise_branch_options"] = self.get_revise_branch_options(script)

        return context


def defaults(config, **kwargs):  # pylint: disable=missing-function-docstring
    base = globals()

    AlembicDashboardView = (  # pylint: disable=invalid-name,redefined-outer-name
        kwargs.get("AlembicDashboardView", base["AlembicDashboardView"])
    )
    AlembicDashboardView.defaults(config)

    AlembicMigrationView = (  # pylint: disable=invalid-name,redefined-outer-name
        kwargs.get("AlembicMigrationView", base["AlembicMigrationView"])
    )
    AlembicMigrationView.defaults(config)


def includeme(config):  # pylint: disable=missing-function-docstring
    defaults(config)



# -*- coding: utf-8; -*-

from unittest.mock import patch

import colander

from wuttaweb.views import common as mod
from wuttaweb.testing import WebTestCase
from wuttaweb.app import establish_theme


class TestCommonView(WebTestCase):

    def make_view(self):
        return mod.CommonView(self.request)

    def test_includeme(self):
        self.pyramid_config.include("wuttaweb.views.common")

    def test_forbidden_view(self):
        view = self.make_view()
        context = view.forbidden_view()
        self.assertEqual(context["index_title"], self.app.get_title())

    def test_notfound_view(self):
        view = self.make_view()
        context = view.notfound_view()
        self.assertEqual(context["index_title"], self.app.get_title())

    def test_home(self):
        self.pyramid_config.add_route("setup", "/setup")
        self.pyramid_config.add_route("login", "/login")
        model = self.app.model
        view = self.make_view()

        # if no users then home page will redirect
        response = view.home(session=self.session)
        self.assertEqual(response.status_code, 302)

        # so add a user
        user = model.User(username="foo")
        self.session.add(user)
        self.session.commit()

        # now we see the home page
        context = view.home(session=self.session)
        self.assertEqual(context["index_title"], self.app.get_title())

        # but if configured, anons will be redirected to login
        self.config.setdefault("wuttaweb.home_redirect_to_login", "true")
        response = view.home(session=self.session)
        self.assertEqual(response.status_code, 302)

        # now only an auth'ed user can see home page
        self.request.user = user
        context = view.home(session=self.session)
        self.assertEqual(context["index_title"], self.app.get_title())

    def test_feedback_make_schema(self):
        view = self.make_view()
        schema = view.feedback_make_schema()
        self.assertIsInstance(schema, colander.Schema)
        self.assertIn("message", schema)

    def test_feedback(self):
        self.pyramid_config.add_route("users.view", "/users/{uuid}")
        model = self.app.model
        user = model.User(username="barney")
        self.session.add(user)
        self.session.commit()

        view = self.make_view()
        with patch.object(view, "feedback_send") as feedback_send:

            # basic send, no user
            self.request.client_addr = "127.0.0.1"
            self.request.method = "POST"
            self.request.POST = {
                "referrer": "/foo",
                "user_name": "Barney Rubble",
                "message": "hello world",
            }
            context = view.feedback()
            self.assertEqual(context, {"ok": True})
            feedback_send.assert_called_once()

            # reset
            feedback_send.reset_mock()

            # basic send, with user
            self.request.user = user
            self.request.POST["user_uuid"] = str(user.uuid)
            with patch.object(mod, "Session", return_value=self.session):
                context = view.feedback()
            self.assertEqual(context, {"ok": True})
            feedback_send.assert_called_once()

            # reset
            self.request.user = None
            feedback_send.reset_mock()

            # invalid form data
            self.request.POST = {"message": "hello world"}
            context = view.feedback()
            self.assertEqual(list(context), ["error"])
            self.assertIn("Required", context["error"])
            feedback_send.assert_not_called()

            # error on send
            self.request.POST = {
                "referrer": "/foo",
                "user_name": "Barney Rubble",
                "message": "hello world",
            }
            feedback_send.side_effect = RuntimeError
            context = view.feedback()
            feedback_send.assert_called_once()
            self.assertEqual(list(context), ["error"])
            self.assertIn("RuntimeError", context["error"])

    def test_feedback_send(self):
        view = self.make_view()
        with patch.object(self.app, "send_email") as send_email:
            view.feedback_send({"user_name": "Barney", "message": "hello world"})
            send_email.assert_called_once_with(
                "feedback", {"user_name": "Barney", "message": "hello world"}
            )

    def test_setup(self):
        self.pyramid_config.add_route("home", "/")
        self.pyramid_config.add_route("login", "/login")
        model = self.app.model
        auth = self.app.get_auth_handler()
        view = self.make_view()

        # at first, can see the setup page
        self.assertEqual(self.session.query(model.User).count(), 0)
        context = view.setup(session=self.session)
        self.assertEqual(context["index_title"], self.app.get_title())

        # so add a user
        user = model.User(username="foo")
        self.session.add(user)
        self.session.commit()

        # once user exists it will always redirect
        response = view.setup(session=self.session)
        self.assertEqual(response.status_code, 302)

        # delete that user
        self.session.delete(user)
        self.session.commit()

        # so we can see the setup page again
        context = view.setup(session=self.session)
        self.assertEqual(context["index_title"], self.app.get_title())

        # and finally, post data to create admin user
        self.request.method = "POST"
        self.request.POST = {
            "username": "barney",
            "__start__": "password:mapping",
            "password": "testpass",
            "password-confirm": "testpass",
            "__end__": "password:mapping",
            "first_name": "Barney",
            "last_name": "Rubble",
        }
        response = view.setup(session=self.session)
        # nb. redirects on success
        self.assertEqual(response.status_code, 302)
        barney = self.session.query(model.User).one()
        self.assertEqual(barney.username, "barney")
        self.assertTrue(auth.check_user_password(barney, "testpass"))
        admin = auth.get_role_administrator(self.session)
        self.assertIn(admin, barney.roles)
        self.assertIsNotNone(barney.person)
        person = barney.person
        self.assertEqual(person.first_name, "Barney")
        self.assertEqual(person.last_name, "Rubble")
        self.assertEqual(person.full_name, "Barney Rubble")

    def test_change_theme(self):
        self.pyramid_config.add_route("home", "/")
        settings = self.request.registry.settings
        establish_theme(settings)
        view = self.make_view()

        # theme is not changed if not provided by caller
        self.assertEqual(settings["wuttaweb.theme"], "default")
        with patch.object(mod, "set_app_theme") as set_app_theme:
            view.change_theme()
            set_app_theme.assert_not_called()
        self.assertEqual(settings["wuttaweb.theme"], "default")

        # but theme will change if provided
        with patch.object(self.request, "params", new={"theme": "butterfly"}):
            with patch.object(mod, "Session", return_value=self.session):
                view.change_theme()
        self.assertEqual(settings["wuttaweb.theme"], "butterfly")

        # flash error if invalid theme is provided
        self.assertFalse(self.request.session.peek_flash("error"))
        with patch.object(self.request, "params", new={"theme": "anotherone"}):
            with patch.object(mod, "Session", return_value=self.session):
                view.change_theme()
        self.assertEqual(settings["wuttaweb.theme"], "butterfly")
        self.assertTrue(self.request.session.peek_flash("error"))
        messages = self.request.session.pop_flash("error")
        self.assertIn("Failed to set theme", messages[0])



# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Report Views
"""

import datetime
import logging
import os
import tempfile

import deform

from wuttaweb.views import MasterView


log = logging.getLogger(__name__)


class ReportView(MasterView):  # pylint: disable=abstract-method
    """
    Master view for :term:`reports <report>`; route prefix is
    ``reports``.

    Notable URLs provided by this class:

    * ``/reports/``
    * ``/reports/XXX``
    """

    model_name = "report"
    model_title = "Report"
    model_key = "report_key"
    filterable = False
    sort_on_backend = False
    creatable = False
    editable = False
    deletable = False
    route_prefix = "reports"
    template_prefix = "/reports"

    grid_columns = [
        "report_title",
        "help_text",
        "report_key",
    ]

    form_fields = [
        "help_text",
    ]

    def __init__(self, request, context=None):
        super().__init__(request, context=context)
        self.report_handler = self.app.get_report_handler()

    def get_grid_data(  # pylint: disable=empty-docstring
        self, columns=None, session=None
    ):
        """ """
        data = []
        for report in self.report_handler.get_reports().values():
            data.append(self.normalize_report(report))
        return data

    def normalize_report(self, report):  # pylint: disable=empty-docstring
        """ """
        return {
            "report_key": report.report_key,
            "report_title": report.report_title,
            "help_text": report.__doc__,
        }

    def configure_grid(self, grid):  # pylint: disable=empty-docstring
        """ """
        g = grid
        super().configure_grid(g)

        # report_key
        g.set_link("report_key")

        # report_title
        g.set_link("report_title")
        g.set_searchable("report_title")

        # help_text
        g.set_searchable("help_text")

    def get_instance(  # pylint: disable=empty-docstring,arguments-differ,unused-argument
        self, **kwargs
    ):
        """ """
        key = self.request.matchdict["report_key"]
        report = self.report_handler.get_report(key)
        if report:
            return self.normalize_report(report)

        raise self.notfound()

    def get_instance_title(self, instance):  # pylint: disable=empty-docstring
        """ """
        report = instance
        return report["report_title"]

    def view(self):
        """
        This lets user "view" the report but in this context that
        means showing them a form with report params, so they can run
        it.
        """
        key = self.request.matchdict["report_key"]
        report = self.report_handler.get_report(key)
        normal = self.normalize_report(report)

        report_url = self.get_action_url("view", normal)
        form = self.make_model_form(
            normal,
            action_method="get",
            action_url=report_url,
            cancel_url=self.get_index_url(),
            show_button_reset=True,
            reset_url=report_url,
            button_label_submit="Run Report",
            button_icon_submit="arrow-circle-right",
        )

        context = {
            "instance": normal,
            "report": report,
            "form": form,
            "xref_buttons": self.get_xref_buttons(report),
        }

        if self.request.GET:
            form.show_button_cancel = False
            context = self.run_report(report, context)

        return self.render_to_response("view", context)

    def configure_form(self, form):  # pylint: disable=empty-docstring
        """ """
        f = form
        super().configure_form(f)
        key = self.request.matchdict["report_key"]
        report = self.report_handler.get_report(key)

        # help_text
        f.set_readonly("help_text")

        # add widget fields for all report params
        schema = f.get_schema()
        report.add_params(schema)
        f.set_fields([node.name for node in schema.children])

    def run_report(self, report, context):
        """
        Run the given report and update view template context.

        This is called automatically from :meth:`view()`.

        :param report:
           :class:`~wuttjamaican:wuttjamaican.reports.Report` instance
           to run.

        :param context: Current view template context.

        :returns: Final view template context.
        """
        form = context["form"]
        controls = list(self.request.GET.items())

        # TODO: must re-inject help_text value for some reason,
        # otherwise its absence screws things up.  why?
        controls.append(("help_text", report.__doc__))

        dform = form.get_deform()
        try:
            params = dform.validate(controls)
        except deform.ValidationFailure:
            log.debug("form not valid: %s", dform.error)
            return context

        data = self.report_handler.make_report_data(report, params)

        columns = self.normalize_columns(report.get_output_columns())
        context["report_columns"] = columns

        format_cols = [col for col in columns if col.get("formatter")]
        if format_cols:
            for record in data["data"]:
                for column in format_cols:
                    if column["name"] in record:
                        value = record[column["name"]]
                        record[column["name"]] = column["formatter"](value)

        params.pop("help_text")
        context["report_params"] = params
        context["report_data"] = data
        context["report_generated"] = datetime.datetime.now()
        return context

    def normalize_columns(self, columns):  # pylint: disable=empty-docstring
        """ """
        normal = []
        for column in columns:
            if isinstance(column, str):
                column = {"name": column}
            column.setdefault("label", column["name"])
            normal.append(column)
        return normal

    def get_download_data(self):  # pylint: disable=empty-docstring
        """ """
        key = self.request.matchdict["report_key"]
        report = self.report_handler.get_report(key)
        params = dict(self.request.GET)
        columns = self.normalize_columns(report.get_output_columns())
        data = self.report_handler.make_report_data(report, params)
        return params, columns, data

    def get_download_path(self, data, ext):  # pylint: disable=empty-docstring
        """ """
        tempdir = tempfile.mkdtemp()
        filename = f"{data['output_title']}.{ext}"
        return os.path.join(tempdir, filename)

    @classmethod
    def defaults(cls, config):  # pylint: disable=empty-docstring
        """ """
        cls._defaults(config)
        cls._report_defaults(config)

    @classmethod
    def _report_defaults(cls, config):
        permission_prefix = cls.get_permission_prefix()
        model_title = cls.get_model_title()

        # overwrite title for "view" perm since it also implies "run"
        config.add_wutta_permission(
            permission_prefix, f"{permission_prefix}.view", f"View / run {model_title}"
        )

        # separate permission to download report files
        config.add_wutta_permission(
            permission_prefix,
            f"{permission_prefix}.download",
            f"Download {model_title}",
        )


def defaults(config, **kwargs):  # pylint: disable=missing-function-docstring
    base = globals()

    ReportView = kwargs.get(  # pylint: disable=invalid-name,redefined-outer-name
        "ReportView", base["ReportView"]
    )
    ReportView.defaults(config)


def includeme(config):  # pylint: disable=missing-function-docstring
    defaults(config)
